/*
 * LMDB Abstraction Layer.
 * Copyright (C) 2023  Yury Gubich <blue@macaw.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LMDBAL_CURSOR_H
#define LMDBAL_CURSOR_H

#include <string>

#include "lmdb.h"
#include "base.h"
#include "storage.h"

namespace LMDBAL {

template <class K, class V>
class Cursor {
    friend class Storage<K, V>;
private:
    enum State {            /**<Cursor state:*/
        closed,             /**<    - closed*/
        openedPublic,       /**<    - opened with public transaction, all storages will be notified about it after it's done*/
        openedPrivate       /**<    - opened with private transaction, only current storage will be notified when cursor is closed*/
    };

    Cursor(Storage<K, V>* parent);
    ~Cursor();

public:
    void open() const;
    void open(TransactionID txn) const;
    void renew() const;
    void renew(TransactionID txn) const;
    void close() const;
    bool opened() const;

    std::pair<K, V> first() const;
    std::pair<K, V> last() const;
    std::pair<K, V> next() const;
    std::pair<K, V> prev() const;
    std::pair<K, V> current() const;

    void first(K& key, V& value) const;
    void last(K& key, V& value) const;
    void next(K& key, V& value) const;
    void prev(K& key, V& value) const;
    void current(K& key, V& value) const;

private:
    void terminated() const;
    void operateCursorRead(K& key, V& value, MDB_cursor_op operation, const std::string& methodName, const std::string& operationName) const;

private:
    Storage<K, V>* storage;
    mutable MDB_cursor* cursor;
    mutable State state;

    inline static const std::string openCursorMethodName = "Cursor::open";      /**<\brief member function name, just for exceptions*/
    inline static const std::string closeCursorMethodName = "Cursor::close";    /**<\brief member function name, just for exceptions*/
    inline static const std::string renewCursorMethodName = "Cursor::renew";    /**<\brief member function name, just for exceptions*/

    inline static const std::string firstMethodName = "first";                  /**<\brief member function name, just for exceptions*/
    inline static const std::string lastMethodName = "last";                    /**<\brief member function name, just for exceptions*/
    inline static const std::string nextMethodName = "next";                    /**<\brief member function name, just for exceptions*/
    inline static const std::string prevMethodName = "prev";                    /**<\brief member function name, just for exceptions*/
    inline static const std::string currentMethodName = "current";              /**<\brief member function name, just for exceptions*/

    inline static const std::string firstOperationName = "Cursor::first";       /**<\brief member function name, just for exceptions*/
    inline static const std::string lastOperationName = "Cursor::last";         /**<\brief member function name, just for exceptions*/
    inline static const std::string nextOperationName = "Cursor::next";         /**<\brief member function name, just for exceptions*/
    inline static const std::string prevOperationName = "Cursor::prev";         /**<\brief member function name, just for exceptions*/
    inline static const std::string currentOperationName = "Cursor::current";   /**<\brief member function name, just for exceptions*/
};

};


#include "cursor.hpp"

#endif //LMDBAL_CURSOR_H
