#include <gtest/gtest.h>

#include "base.h"
#include "storage.h"
#include "cache.h"
#include "cursor.h"

class CacheCursorTest : public ::testing::Test {
protected:
    CacheCursorTest():
        ::testing::Test(),
        cache (db->getCache<uint64_t, std::string>("table1")),
        emptyCache (db->getCache<uint64_t, std::string>("empty")) {}

    ~CacheCursorTest() {}

    static void SetUpTestSuite() {
        if (db == nullptr) {
            db = new LMDBAL::Base("testBase");
            db->addCache<uint64_t, std::string>("table1");
            db->addCache<uint64_t, std::string>("empty");
            db->open();
        }
    }

    static void TearDownTestSuite() {
        db->close();
        db->removeDirectory();
        delete db;
        db = nullptr;
        cursor = nullptr;
        emptyCursor = nullptr;
    }

    static LMDBAL::Base* db;
    static LMDBAL::Cursor<uint64_t, std::string>* cursor;
    static LMDBAL::Cursor<uint64_t, std::string>* emptyCursor;
    static LMDBAL::TransactionID transaction;

    LMDBAL::Cache<uint64_t, std::string>* cache;
    LMDBAL::Cache<uint64_t, std::string>* emptyCache;
};

LMDBAL::Base* CacheCursorTest::db = nullptr;
LMDBAL::Cursor<uint64_t, std::string>* CacheCursorTest::cursor = nullptr;
LMDBAL::Cursor<uint64_t, std::string>* CacheCursorTest::emptyCursor = nullptr;
LMDBAL::TransactionID CacheCursorTest::transaction = nullptr;

static const std::map<uint64_t, std::string> data({
    {245665783, "bothering nerds"},
    {3458, "resilent pick forefront"},
    {105190, "apportunity legal bat"},
    {6510, "outside"},
    {7438537, "damocles plush apparently rusty"},
    {19373572, "local guidence"},
    {138842, "forgetting tusks prepare"},
    {981874, "butchered soaking pawn"},
    {19302, "tanned inmate"},
    {178239, "custody speaks neurotic"},
});

TEST_F(CacheCursorTest, PopulatingTheTable) {
    uint32_t amount = cache->addRecords(data);
    EXPECT_EQ(amount, data.size());
}

TEST_F(CacheCursorTest, Creation) {
    cursor = cache->createCursor();
    emptyCursor = emptyCache->createCursor();

    EXPECT_THROW(cursor->first(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->last(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->next(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->prev(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->current(), LMDBAL::CursorNotReady);

    cursor->open();
}

TEST_F(CacheCursorTest, FirstPrivate) {
    EXPECT_EQ(cache->count(), data.size());

    std::pair<uint64_t, std::string> element = cursor->first();
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, NextPrivate) {
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    reference++;
    for (; reference != data.end(); ++reference) {
        std::pair<uint64_t, std::string> element = cursor->next();
        EXPECT_EQ(element.first, reference->first);
        EXPECT_EQ(element.second, reference->second);
        EXPECT_EQ(cache->count(), data.size());
    }

    EXPECT_THROW(cursor->next(), LMDBAL::NotFound);
    EXPECT_EQ(cache->count(), data.size());

    std::pair<uint64_t, std::string> element = cursor->first();
    reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, LastPrivate) {
    EXPECT_EQ(cache->count(), data.size());

    std::pair<uint64_t, std::string> element = cursor->last();
    std::map<uint64_t, std::string>::const_reverse_iterator reference = data.rbegin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, PrevPrivate) {
    std::map<uint64_t, std::string>::const_reverse_iterator reference = data.rbegin();

    reference++;
    for (; reference != data.rend(); ++reference) {
        std::pair<uint64_t, std::string> element = cursor->prev();
        EXPECT_EQ(element.first, reference->first);
        EXPECT_EQ(element.second, reference->second);
        EXPECT_EQ(cache->count(), data.size());
    }

    EXPECT_THROW(cursor->prev(), LMDBAL::NotFound);
    EXPECT_EQ(cache->count(), data.size());

    std::pair<uint64_t, std::string> element = cursor->last();
    reference = data.rbegin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, CurrentPrivate) {
    std::pair<uint64_t, std::string> element = cursor->first();
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    element = cursor->current();
    EXPECT_EQ(cache->count(), data.size());

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    cursor->next();
    element = cursor->current();
    ++reference;

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    cursor->next();
    cursor->next();
    cursor->prev();
    element = cursor->current();
    ++reference;
    ++reference;
    --reference;

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, Destruction) {
    cursor->close();

    EXPECT_THROW(cursor->first(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->last(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->next(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->prev(), LMDBAL::CursorNotReady);
    EXPECT_THROW(cursor->current(), LMDBAL::CursorNotReady);

    EXPECT_THROW(emptyCache->destroyCursor(cursor), LMDBAL::Unknown);
    cache->destroyCursor(cursor);

    cursor = cache->createCursor();
}

TEST_F(CacheCursorTest, FirstPublic) {
    transaction = db->beginTransaction();

    cursor->open(transaction);
    std::pair<uint64_t, std::string> element = cursor->first();
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, NextPublic) {
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    reference++;
    for (; reference != data.end(); ++reference) {
        std::pair<uint64_t, std::string> element = cursor->next();
        EXPECT_EQ(element.first, reference->first);
        EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
    }

    EXPECT_THROW(cursor->next(), LMDBAL::NotFound);
    EXPECT_EQ(cache->count(), data.size());

    std::pair<uint64_t, std::string> element = cursor->first();
    reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
}

TEST_F(CacheCursorTest, LastPublic) {
    std::pair<uint64_t, std::string> element = cursor->last();
    std::map<uint64_t, std::string>::const_reverse_iterator reference = data.rbegin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, PrevPublic) {
    std::map<uint64_t, std::string>::const_reverse_iterator reference = data.rbegin();

    reference++;
    for (; reference != data.rend(); ++reference) {
        std::pair<uint64_t, std::string> element = cursor->prev();
        EXPECT_EQ(element.first, reference->first);
        EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
    }

    EXPECT_THROW(cursor->prev(), LMDBAL::NotFound);

    std::pair<uint64_t, std::string> element = cursor->last();
    reference = data.rbegin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, CurrentPublic) {
    std::pair<uint64_t, std::string> element = cursor->first();
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    element = cursor->current();
    EXPECT_EQ(cache->count(), data.size());

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    cursor->next();
    element = cursor->current();
    ++reference;

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);

    cursor->next();
    cursor->next();
    cursor->prev();
    element = cursor->current();
    ++reference;
    ++reference;
    --reference;

    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_EQ(cache->count(), data.size());
}

TEST_F(CacheCursorTest, CornerCases) {
    db->abortTransaction(transaction);
    EXPECT_THROW(cursor->current(), LMDBAL::Unknown);
    cursor->close();

    emptyCursor->open();
    EXPECT_THROW(emptyCursor->first(), LMDBAL::NotFound);
    EXPECT_THROW(emptyCursor->last(), LMDBAL::NotFound);
    EXPECT_THROW(emptyCursor->next(), LMDBAL::NotFound);
    EXPECT_THROW(emptyCursor->prev(), LMDBAL::NotFound);
    EXPECT_THROW(emptyCursor->current(), LMDBAL::Unknown);
    emptyCursor->close();

    cursor->open();
    EXPECT_THROW(cursor->current(), LMDBAL::Unknown);       //yeah, nice thing to write in the doc

    std::map<uint64_t, std::string>::const_reverse_iterator breference = data.rbegin();
    std::pair<uint64_t, std::string> element(cursor->prev());
    EXPECT_EQ(element.first, breference->first);            //nice thing to write in the doc, again!
    EXPECT_EQ(element.second, breference->second);
    element = cursor->current();
    EXPECT_EQ(element.first, breference->first);
    EXPECT_EQ(element.second, breference->second);
    EXPECT_THROW(cursor->next(), LMDBAL::NotFound);
    cursor->close();

    cursor->open();
    element = cursor->next();
    std::map<uint64_t, std::string>::const_iterator reference = data.begin();
    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    element = cursor->current();
    EXPECT_EQ(element.first, reference->first);
    EXPECT_EQ(element.second, reference->second);
    EXPECT_THROW(cursor->prev(), LMDBAL::NotFound);
}

